# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""


import tempfile

import pytest


@pytest.fixture()
def tmppath():
    """Application fixture."""
    tmppath = tempfile.mkdtemp()
    yield tmppath
    return tmppath
